// code by jph

// * default formats encouraged: GL_T2F_N3F_V3F, or GL_T2F_C4F_N3F_V3F
// * shadow volumes restricted to models with less than 65536 vertices

// credits: shadow volumes after the paper by Cass Everitt and Mark J. Kildard,
//          Practical and Robust Stenciled Shadow Volumes for Hardware-Accelerated Rendering

class meshgl : public mesh<float> {
public:  
  meshgl();
  meshgl(char *fn);
  
  static meshgl surface(const tensor<float>& val,tensor<double> ext);
  
  int offsetT2F(void);
  int offsetC3F(void);
  int offsetN3F(void);
  int offsetV3F(void);
  
  meshgl map(tensor<double> mat);
  void setC3F(tensor<double> col);
  void setC4F(tensor<double> col);
  tensor<float> tria_N3F(void);
  tensor<float> quad_N3F(void);
  void updateN3F(void);

  void draw(void);
  void shadow(tensor<double> rel);

  void drawN3F(float sca,int skip);

  double          volume(void);
  tensor<double>  inertia(void);
  tensor<double>  centroid(void);

  char* toInfo(void);
  
};

meshgl::meshgl() : mesh<float>() {}

meshgl::meshgl(char* fn) : mesh<float>(fn) {
  assert(mesh<float>::valid);
  if (arr.numel)
    switch (arr.size[0]) {
      case  3: format=GL_V3F;             break;
      case  5: format=GL_T2F_V3F;         break;
      case  8: format=GL_T2F_N3F_V3F;     break; // GL_T2F_C3F_V3F is ignored
      case 10: format=GL_C4F_N3F_V3F;     break;
      case 12: format=GL_T2F_C4F_N3F_V3F; break;
      case 15: format=GL_T4F_C4F_N3F_V4F; break;
      default:
        assert(false);
    }
}

meshgl meshgl::surface(const tensor<float>& val,tensor<double> ext) {
  int c0,c1,c2=0,c3=0;
  meshgl ret;
  ret.ind[MESH_QUAD]=tensor<uint>(2,MESH_COMP[MESH_QUAD],(val.size[0]-1)*(val.size[1]-1));
  ret.arr=tensor<float>(2,8,val.numel);
  ret.format=GL_T2F_N3F_V3F;
  for (c1=0;c1<val.size[1];++c1)
  for (c0=0;c0<val.size[0];++c0) {
    ret.arr(5,c2)=c0*ext(0)/(val.size[0]-1);
    ret.arr(6,c2)=c1*ext(1)/(val.size[1]-1);
    ret.arr(7,c2)=val.data[c0+c1*val.size[0]];
    if (c0&&c1)
      ret.ind[MESH_QUAD](0,c3++)=c2;
    ++c2;
  }
  assert(ret.ind[MESH_QUAD].size[1]==c3);
  for (c0=0;c0<ret.ind[MESH_QUAD].size[1];++c0) {
    ret.ind[MESH_QUAD](1,c0)=ret.ind[MESH_QUAD](0,c0)-1;
    ret.ind[MESH_QUAD](2,c0)=ret.ind[MESH_QUAD](0,c0)-val.size[0]-1;
    ret.ind[MESH_QUAD](3,c0)=ret.ind[MESH_QUAD](0,c0)-val.size[0];
  }
  for (c0=0;c0<ret.ind[MESH_QUAD].numel;++c0)
    assert(0<=ret.ind[MESH_QUAD].data[c0]&&ret.ind[MESH_QUAD].data[c0]<ret.arr.size[1]);
  ret.updateN3F();
  return ret;
}

int meshgl::offsetT2F(void) {
  switch (format) {
    case GL_T2F_V3F:          return 0;
    case GL_T2F_C4UB_V3F:     return 0;
    case GL_T2F_C3F_V3F:      return 0;
    case GL_T2F_N3F_V3F:      return 0;
    case GL_T2F_C4F_N3F_V3F:  return 0;
    case GL_T4F_C4F_N3F_V4F:  return 0;
  }
  assert(false);
  return 0;
}

int meshgl::offsetC3F(void) {
  switch (format) {
    case GL_C3F_V3F:          return 0;
    case GL_C4F_N3F_V3F:      return 0;
    case GL_T2F_C3F_V3F:      return 2;
    case GL_T2F_C4F_N3F_V3F:  return 2;
    case GL_T4F_C4F_N3F_V4F:  return 4;
  }
  assert(false);
  return 0;
}

int meshgl::offsetN3F(void) {
  switch (format) {
    case GL_N3F_V3F:          return 0;
    case GL_C4F_N3F_V3F:      return 4;
    case GL_T2F_N3F_V3F:      return 2;
    case GL_T2F_C4F_N3F_V3F:  return 6;
    case GL_T4F_C4F_N3F_V4F:  return 8;
  }
  assert(false);
  return 0;
}

int meshgl::offsetV3F(void) {
  switch (format) {
    case GL_V3F:              return 0;
    case GL_C3F_V3F:          return 3;
    case GL_N3F_V3F:          return 3;
    case GL_C4F_N3F_V3F:      return 7;
    case GL_T2F_V3F:          return 2;
    case GL_T2F_C4UB_V3F:     return 3;
    case GL_T2F_C3F_V3F:      return 5;
    case GL_T2F_N3F_V3F:      return 5;
    case GL_T2F_C4F_N3F_V3F:  return 9;
    case GL_T4F_C4F_N3F_V4F:  return 11;
  }
  assert(false);
  return 0;
}

// applies 4x4 affine transformation mat to V3F strip
meshgl meshgl::map(tensor<double> mat) {
  meshgl ret(*this);
  tensor<float> vec(1,4);
  vec(3)=1;
  for (int c0=offsetV3F();c0<arr.numel;c0+=arr.size[0]) {
    vec.in(0,&arr.data[c0],3);
    ret.arr.in(c0,(mat*vec.to<double>()).to<float>().data,3);
  }
  ret.updateN3F();
  return ret;
}

void meshgl::setC3F(tensor<double> col) {
  tensor<float> cof=col.to<float>();
  for (int c0=offsetC3F();c0<arr.numel;c0+=arr.size[0])
    arr.in(c0,cof.data,3);
}

void meshgl::setC4F(tensor<double> col) {
  tensor<float> cof=col.to<float>();
  for (int c0=offsetC3F();c0<arr.numel;c0+=arr.size[0])
    arr.in(c0,cof.data,4);
}

// normals per triangle
tensor<float> meshgl::tria_N3F(void) {
  int v3f=offsetV3F();
  tensor<float> v0(1,3),v1(1,3),v2(1,3);
  tensor<float> ret(2,3,ind[MESH_TRIA].size[1]);
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    v0.in(0,&arr(v3f,ind[MESH_TRIA].data[0+c0]),3);
    v1.in(0,&arr(v3f,ind[MESH_TRIA].data[1+c0]),3);
    v2.in(0,&arr(v3f,ind[MESH_TRIA].data[2+c0]),3);
    ret.in(c0,(v1-v0).x(v2-v0).normal().data,3);    
  }
  return ret;
}

// normals per quad
// assumes quads are flat
tensor<float> meshgl::quad_N3F(void) {
  int v3f=offsetV3F();
  tensor<float> v0(1,3),v1(1,3),v2(1,3);
  tensor<float> ret(2,3,ind[MESH_QUAD].size[1]);
  for (int c0=0;c0<ind[MESH_QUAD].size[1];++c0) {
    v0.in(0,&arr(v3f,ind[MESH_QUAD].data[0+c0*4]),3);
    v1.in(0,&arr(v3f,ind[MESH_QUAD].data[1+c0*4]),3);
    v2.in(0,&arr(v3f,ind[MESH_QUAD].data[2+c0*4]),3);
    ret.in(c0*3,(v1-v0).x(v2-v0).normal().data,3);
  }
  return ret;
}

//tensor<float> meshgl::tria_D3F(void) {
//  int c0;
//  int t0,t1,t2;
//  int n3f=offsetN3F();
//  int v3f=offsetV3F();
//  tensor<float> v0(1,3),v1(1,3),v2(1,3);
//  tensor<float> ret(2,3,ind[MESH_TRIA].size[1]);
//  for (c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
//    v0.in(0,&arr(v3f,ind[MESH_TRIA].data[0+c0]),3);
//    v1.in(0,&arr(v3f,ind[MESH_TRIA].data[1+c0]),3);
//    v2.in(0,&arr(v3f,ind[MESH_TRIA].data[2+c0]),3);
//    ret.in(c0,(v1-v0).x(v2-v0).normal().data,3);    
//  }
//  return ret;
//}

void meshgl::updateN3F(void) {
  int c0;
  int t0,t1,t2,t3;
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<float> v0(1,3),v1(1,3),v2(1,3);
  for (c0=n3f;c0<arr.numel;c0+=arr.size[0])
    arr.data[c0+0]=arr.data[c0+1]=arr.data[c0+2]=0;
  tensor<float> ret;
  ret=tria_N3F();
  for (c0=0;c0<ind[MESH_TRIA].size[1];++c0) {
    t0=ind[MESH_TRIA](0,c0);
    t1=ind[MESH_TRIA](1,c0);
    t2=ind[MESH_TRIA](2,c0);
    v0.in(0,ret^c0*3,3);
    arr(n3f+0,t0)+=v0(0);
    arr(n3f+1,t0)+=v0(1);
    arr(n3f+2,t0)+=v0(2);
    arr(n3f+0,t1)+=v0(0);
    arr(n3f+1,t1)+=v0(1);
    arr(n3f+2,t1)+=v0(2);
    arr(n3f+0,t2)+=v0(0);
    arr(n3f+1,t2)+=v0(1);
    arr(n3f+2,t2)+=v0(2);
  }
  ret=quad_N3F();
  for (c0=0;c0<ind[MESH_QUAD].size[1];++c0) {
    t0=ind[MESH_QUAD](0,c0);
    t1=ind[MESH_QUAD](1,c0);
    t2=ind[MESH_QUAD](2,c0);
    t3=ind[MESH_QUAD](3,c0);
    v0.in(0,ret^c0*3,3);
    arr(n3f+0,t0)+=v0(0);
    arr(n3f+1,t0)+=v0(1);
    arr(n3f+2,t0)+=v0(2);
    arr(n3f+0,t1)+=v0(0);
    arr(n3f+1,t1)+=v0(1);
    arr(n3f+2,t1)+=v0(2);
    arr(n3f+0,t2)+=v0(0);
    arr(n3f+1,t2)+=v0(1);
    arr(n3f+2,t2)+=v0(2);
    arr(n3f+0,t3)+=v0(0);
    arr(n3f+1,t3)+=v0(1);
    arr(n3f+2,t3)+=v0(2);
  }
  for (c0=n3f;c0<arr.numel;c0+=arr.size[0]) {
    v0.in(0,&arr.data[c0],3);
    v0=v0.normal();
    arr.data[c0+0]=v0(0);
    arr.data[c0+1]=v0(1);
    arr.data[c0+2]=v0(2);
  }
}

void meshgl::draw(void) {
  glPushAttrib(GL_ENABLE_BIT);  
    glMaterialfv(GL_FRONT,GL_SPECULAR,specular.to<float>().data);
    glMaterialf(GL_FRONT,GL_SHININESS,shininess);
	  if (blend) {
	    glEnable(GL_BLEND);
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      //glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);
      //glBlendFunc(GL_ONE,GL_ONE);
    }
    if (texture) {
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,texture);
	  }
    glColor4dv(color.data);
    glInterleavedArrays(format,arr.size[0]*4,arr.data);
    glDrawElements(GL_POINTS    ,ind[MESH_VERT].numel,GL_UNSIGNED_INT,ind[MESH_VERT].data);
    glDrawElements(GL_LINES     ,ind[MESH_EDGE].numel,GL_UNSIGNED_INT,ind[MESH_EDGE].data);
    glDrawElements(GL_TRIANGLES ,ind[MESH_TRIA].numel,GL_UNSIGNED_INT,ind[MESH_TRIA].data);
    glDrawElements(GL_QUADS     ,ind[MESH_QUAD].numel,GL_UNSIGNED_INT,ind[MESH_QUAD].data);
  glPopAttrib();
}

#define MESHGL_SHADOW   1e10

#define MESHGL_SHADOW_NORMAL            \
    vrt.in(0,&arr.data[v3f+t0*sz0],3);  \
    vrt-=pos;                           \
    v0.in(0,&arr.data[v3f+t0*sz0],3);   \
    v1.in(0,&arr.data[v3f+t1*sz0],3);   \
    v2.in(0,&arr.data[v3f+t2*sz0],3);     

#define MESHGL_SHADOW_TRIA1             \
    t0=ind[MESH_TRIA].data[c0  ];       \
    t1=ind[MESH_TRIA].data[c0+1];       \
    t2=ind[MESH_TRIA].data[c0+2];       \
    MESHGL_SHADOW_NORMAL
    
#define MESHGL_SHADOW_TRIA2             \
      blk.in( 0,v0.data,3); v0-=pos; v0*=MESHGL_SHADOW; blk.in( 3,v0.data,3); \
      blk.in( 6,v1.data,3); v1-=pos; v1*=MESHGL_SHADOW; blk.in( 9,v1.data,3); \
      blk.in(12,v2.data,3); v2-=pos; v2*=MESHGL_SHADOW; blk.in(15,v2.data,3); \
      glDrawElements(GL_TRIANGLES,tri.numel,GL_UNSIGNED_INT,tri.data);      

#define MESHGL_SHADOW_QUAD1             \
    t0=ind[MESH_QUAD].data[c0  ];       \
    t1=ind[MESH_QUAD].data[c0+1];       \
    t2=ind[MESH_QUAD].data[c0+2];       \
    MESHGL_SHADOW_NORMAL
    
#define MESHGL_SHADOW_QUAD2             \
      t3=ind[MESH_QUAD].data[c0+3];     \
      v3.in(0,&arr.data[v3f+t3*sz0],3); \
      blk.in( 0,v0.data,3); v0-=pos; v0*=MESHGL_SHADOW; blk.in( 3,v0.data,3); \
      blk.in( 6,v1.data,3); v1-=pos; v1*=MESHGL_SHADOW; blk.in( 9,v1.data,3); \
      blk.in(12,v2.data,3); v2-=pos; v2*=MESHGL_SHADOW; blk.in(15,v2.data,3); \
      blk.in(18,v3.data,3); v3-=pos; v3*=MESHGL_SHADOW; blk.in(21,v3.data,3); \
      glDrawElements(GL_QUADS,qua.numel,GL_UNSIGNED_INT,qua.data);      
    
#define MESHGL_SHADOW_EDGES             \
  for (int c0=0;c0<edg.numel;++c0) {    \
    t1=edg.data[c0]&0xffff;             \
    t0=edg.data[c0]>>16;                \
    vrt.in(0,&arr.data[v3f+t0*sz0],3);  \
    blk.in(0,vrt.data,3);               \
    vrt-=pos; vrt*=MESHGL_SHADOW;       \
    blk.in(3,vrt.data,3);               \
    vrt.in(0,&arr.data[v3f+t1*sz0],3);  \
    blk.in(9,vrt.data,3);               \
    vrt-=pos; vrt*=MESHGL_SHADOW;       \
    blk.in(6,vrt.data,3);               \
    glDrawElements(GL_QUADS,qst.numel,GL_UNSIGNED_INT,qst.data); \
  }

#define MESHGL_SHADOW_EDGE(X,Y)         \
      uori=X; uori<<=16; uori+=Y;       \
      if (X<Y)                          \
        eds.toggle(uori,uori);          \
      else {                            \
        usrt=Y; usrt<<=16; usrt+=X;     \
        eds.toggle(usrt,uori);          \
      }

void meshgl::shadow(tensor<double> rel) {
  assert(arr.size[1]<65536);
  uint t0,t1,t2,t3;
  uint sz0=arr.size[0];
  uint v3f=offsetV3F();
  hash<uint,uint> eds(sqrt((double)ind[MESH_TRIA].size[1])*6,set<uint>::map);
  tensor<float> pos(1,3); pos.in(0,rel.to<float>().data,3);
  tensor<float> vrt(1,3);
  tensor<float> blk(2,3,8);
  tensor<int> tri=tensor<int>::matrix(3,2).as(0,2,4,1,5,3);
  tensor<int> qua=tensor<int>::matrix(4,2).as(0,2,4,6,1,7,5,3);
  uint uori,usrt;
  glInterleavedArrays(GL_V3F,blk.size[0]*4,blk.data);
  glCullFace(GL_FRONT);
  glStencilOp(GL_KEEP,GL_INCR,GL_KEEP);
  tensor<float> v0(1,3),v1(1,3),v2(1,3),v3(1,3);
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    MESHGL_SHADOW_TRIA1
    if (((v1-v0).x(v2-v0)|vrt)<0) {
      MESHGL_SHADOW_TRIA2
      MESHGL_SHADOW_EDGE(t0,t1)
      MESHGL_SHADOW_EDGE(t1,t2)
      MESHGL_SHADOW_EDGE(t2,t0)
    }
  }
  for (int c0=0;c0<ind[MESH_QUAD].numel;c0+=4) {
    MESHGL_SHADOW_QUAD1
    if (((v1-v0).x(v2-v0)|vrt)<0) {
      MESHGL_SHADOW_QUAD2
      MESHGL_SHADOW_EDGE(t0,t1)
      MESHGL_SHADOW_EDGE(t1,t2)
      MESHGL_SHADOW_EDGE(t2,t3)
      MESHGL_SHADOW_EDGE(t3,t0)
    }
  }
  tensor<uint> edg=eds.toTensor();
  tensor<int> qst=tensor<int>::index(1,4);
  MESHGL_SHADOW_EDGES
  glCullFace(GL_BACK);
  glStencilOp(GL_KEEP,GL_DECR,GL_KEEP);
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    MESHGL_SHADOW_TRIA1
    if (((v1-v0).x(v2-v0)|vrt)<0) {
      MESHGL_SHADOW_TRIA2
    }
  }  
  for (int c0=0;c0<ind[MESH_QUAD].numel;c0+=4) {
    MESHGL_SHADOW_QUAD1
    if (((v1-v0).x(v2-v0)|vrt)<0) {      
      MESHGL_SHADOW_QUAD2
    }
  }  
  MESHGL_SHADOW_EDGES  
}

void meshgl::drawN3F(float sca,int skip) {
  skip=arr.size[0]*++skip;
  int c3f=offsetC3F();
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<float> pnt(1,3);
  tensor<float> nrm(1,3);
  glBegin(GL_LINES);
    for (int c0=0;c0<arr.numel;c0+=skip) {
      glColor3fv(&arr.data[c0+c3f]);
      nrm.in(0,&arr.data[c0+n3f],3);
      glNormal3fv(nrm.data);
      glVertex3fv(pnt.in(0,&arr.data[c0+v3f],3).data);
      glVertex3fv((pnt+nrm*sca).data);
    }
  glEnd();
}

double meshgl::volume(void) {
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<float> trh(2,3,3);
  double sum=0;
  for (int c0=0;c0<ind[MESH_TRIA].size[1];++c0) {
    trh.in(0,&arr(v3f,ind[MESH_TRIA](0,c0)),3);
    trh.in(3,&arr(v3f,ind[MESH_TRIA](1,c0)),3);
    trh.in(6,&arr(v3f,ind[MESH_TRIA](2,c0)),3);
    sum+=trh.to<double>().det();
  }
  for (int c0=0;c0<ind[MESH_QUAD].size[1];++c0) {
    trh.in(0,&arr(v3f,ind[MESH_QUAD](0,c0)),3);
    trh.in(3,&arr(v3f,ind[MESH_QUAD](1,c0)),3);
    trh.in(6,&arr(v3f,ind[MESH_QUAD](2,c0)),3);
    sum+=trh.to<double>().det();
    trh.in(3,&arr(v3f,ind[MESH_QUAD](3,c0)),3);
    sum-=trh.to<double>().det();
  }
  return sum/6;
}

tensor<double> meshgl::inertia(void) {
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<double> ine=tensor<double>::matrix(3,3).as( 2.0, 1.0, 1.0,  1.0, 2.0, 1.0,  1.0, 1.0, 2.0);
  ine*=1.0/120;
  tensor<float> trh(2,3,3);
  tensor<double> trd;
  tensor<double> buf(2,3,3);
  double vol=0,det;
  for (int c0=0;c0<ind[MESH_TRIA].size[1];++c0) {
    trh.in(0,&arr(v3f,ind[MESH_TRIA](0,c0)),3);
    trh.in(3,&arr(v3f,ind[MESH_TRIA](1,c0)),3);
    trh.in(6,&arr(v3f,ind[MESH_TRIA](2,c0)),3);
    trd=trh.to<double>();
    det=trd.det();
    vol+=det;
    buf+=(trd*ine*trd.transpose())*det;
  }
  buf=tensor<double>::eye(3)*buf.trace(0,1).data[0]-buf;
  tensor<double> ret(2,4,4);
  ret.in(0,&buf.data[0],3);
  ret.in(4,&buf.data[3],3);
  ret.in(8,&buf.data[6],3);
  ret.data[15]=vol/6;
  return ret;
}


tensor<double> meshgl::centroid(void) {
  int v3f=offsetV3F();
  tensor<float> trh(2,3,3);
  tensor<double> trd;
  tensor<double> ret(1,4);
  tensor<double> add(1,4); add(3)=4/3.0;
  for (int c0=0;c0<ind[MESH_TRIA].size[1];++c0) {
    trh.in(0,&arr(v3f,ind[MESH_TRIA](0,c0)),3);
    trh.in(3,&arr(v3f,ind[MESH_TRIA](1,c0)),3);
    trh.in(6,&arr(v3f,ind[MESH_TRIA](2,c0)),3);
    trd=trh.to<double>();
    add.in(0,trd.mean().data,3);
    ret+=add*trd.det();
  }
  return ret/ret(3);
}

char* meshgl::toInfo(void) {
  sprintf(MESH_BUFFER,"%s V3F:%i N3F:%i",mesh<float>::toInfo(),offsetV3F(),offsetN3F());
  return MESH_BUFFER;
}
